"""
Author: Cory Chu <cory@gwlab.page>
"""

import numpy as np
from scipy import integrate
from read_rapidpe.transform import transform_mceta_to_m1m2
from read_rapidpe.transform import jacobian_m1m2_by_mceta


def fixed_dblquad(f, xa, xb, ya, yb, order=500):
    """
    Fixed-order dblquad
    Ref: https://stackoverflow.com/questions/42272817/is-it-possible-to-compute-double-integral-using-scipy-integrate-fixed-quad

    Parameters
    ----------
    f(y, x) : callable
        y must be the first argument and x the second argument.
    xa, xb : float
        The limits of integration in x: xa < xb
    ya, yb : float
        The limits of integration in y: ya < yb

    """  # noqa: E501

    def f_of_x(xs, n):
        return np.array(
            [integrate.fixed_quad(f, ya, yb, args=(x,), n=n)[0]
                for x in np.array(xs)]
            )

    return integrate.fixed_quad(f_of_x, xa, xb, n=order, args=(order,))


class Uniform_in_m1m2(object):
    def __init__(self, m_lower, m_upper):
        self.m_lower = m_lower
        self.m_upper = m_upper
        self.ans = 2. / (self.m_upper - self.m_lower)**2
        # self._prior_np = np.vectorize(self._prior, excluded="self")

    def __call__(self, m1, m2):
        if np.isscalar(m1):
            return self._prior(m1, m2)
        else:
            return self._prior_np(m1, m2)

    def _prior_np(self, m1, m2):
        return (np.piecewise(
                    m1, [m1 < m2, m1 >= m2], [0., 1.]) *
                np.piecewise(
                    m2, [m2 < self.m_lower, m2 >= self.m_lower], [0., 1.]) *
                np.piecewise(
                    m1, [m1 <= self.m_upper, m1 > self.m_upper], [1., 0.]) *
                self.ans)

    def _prior(self, m1, m2):
        if m1 < m2:
            return 0.
        elif m2 < self.m_lower or m1 > self.m_upper:
            return 0.
        else:
            return self.ans


class Uniform_in_m1m2_nsbh(object):
    def __init__(self, m_lower, m_mid, m_upper):
        self.m_lower = m_lower
        self.m_mid = m_mid
        self.m_upper = m_upper
        self._prior_np = np.vectorize(self._prior, excluded="self")

    def __call__(self, m1, m2):
        return self._prior_np(m1, m2)

    def _prior(self, m1, m2):
        if m1 < m2:
            return 0.
        elif m2 < self.m_lower or m1 > self.m_upper:
            return 0.
        elif m2 > self.m_mid or m1 < self.m_mid:
            return 0.
        else:
            return 1./((self.m_upper - self.m_mid)*(self.m_mid - self.m_lower))


def evidence_integral(res, prior):
    N = len(res.samples["mass_1"])
    return np.sum(prior(res.samples["mass_1"], res.samples["mass_2"]))/N


def _Z_in_m1m2(result, method="riemann-sum"):
    """
    calculate Z = \\int L_marg(d | m1, m2) dm1 dm2
    """

    mc_min = result.chirp_mass.min()+0.00001
    mc_max = result.chirp_mass.max()-0.00001
    eta_min = result.symmetric_mass_ratio.min()+0.00001
    eta_max = result.symmetric_mass_ratio.max()-0.00001

    # Choose interpolating likelihood (if not exist)
    if not hasattr(result, "log_likelihood"):
        result.do_interpolate_marg_log_likelihood_m1m2(method="linear-scipy")

    def f(mc, eta):
        mass1, mass2 = transform_mceta_to_m1m2(mc, eta)
        likelihood = np.exp(result.log_likelihood(mass1, mass2))
        jacobian = jacobian_m1m2_by_mceta(mc, eta)
        return likelihood * jacobian

    if method == "riemann-sum":
        # Define Mesh in mc, eta space for likelihood integral
        mclist = np.linspace(mc_min, mc_max, 500)
        etalist = np.linspace(eta_min, eta_max, 500)
        delta_mc = mclist[1]-mclist[0]
        delta_eta = etalist[1]-etalist[0]
        mc, eta = np.meshgrid(mclist, etalist)

        Z = delta_eta * delta_mc * np.sum(f(mc, eta))
        return Z

    elif method == "scipy-dblquad":
        def g(eta, mc):  # g(y, x) for dblquad
            return f(mc, eta)

        Z = integrate.dblquad(g, mc_min, mc_max, eta_min, eta_max,
                              epsrel=1e-2)  # FIXME: Is 1e-2 okay?
        return Z[0]

    elif method == "fixed-dblquad":
        def g(eta, mc):  # g(y, x) for dblquad
            return f(mc, eta)
        Z = fixed_dblquad(g, mc_min, mc_max, eta_min, eta_max, order=500)
        return Z[0]

    else:
        raise ValueError(
            'method= "scipy-dblquad" or "riemann-sum"'
            )


def _bayes_factor(result, prior):
    """
    bayes factor = p(d|H_1) / p(d|H_0)
    """

    if not hasattr(result, "samples"):
        result.generate_samples()

    bayes_factor = _Z_in_m1m2(result) * evidence_integral(result, prior)
    return bayes_factor


def p_astro(result, ml=1, mm=3, mh=100):
    # Define Rate of BNS, NSBH, and BBH
    # rate[\alpha] = P(H_\alpha \ver H_1)
    # FIXME: Replaced by actual rate
    p_bns_in_H1 = (mm - ml)**2/2
    p_nsbh_in_H1 = (mm - ml)*(mh - mm)
    p_bbh_in_H1 = (mh - mm)**2/2

    rate = np.array([
        p_bns_in_H1,
        p_nsbh_in_H1,
        p_bbh_in_H1
    ])
    rate /= rate.sum()

    # Define prior \pi(\theta \vert H_\alpha)
    # FIXME: Replaced by priors from population model
    prior_bns = Uniform_in_m1m2(ml, mm)
    prior_nsbh = Uniform_in_m1m2_nsbh(ml, mm, mh)
    prior_bbh = Uniform_in_m1m2(mm, mh)

    def prior_astro(m1, m2):
        prior = rate[0] * prior_bns(m1, m2) + \
                rate[1] * prior_nsbh(m1, m2) + \
                rate[2] * prior_nsbh(m1, m2)
        return prior

    # Define prior odd, i.e., P(H_1)/P(H_0)
    prior_odd = 1

    # Calculation
    if not hasattr(result, "samples"):
        result.generate_samples()

    p_astro = np.array([
        evidence_integral(result, prior_bns),
        evidence_integral(result, prior_nsbh),
        evidence_integral(result, prior_bbh),
    ])
    p_astro *= rate
    p_astro /= p_astro.sum()

    p_terr = 1/(1 + _bayes_factor(result, prior_astro) * prior_odd)
    p_astro *= (1-p_terr)

    p_astro = {
        "BNS": p_astro[0],
        "NSBH": p_astro[1],
        "BBH": p_astro[2],
        "Terrestrial": p_terr,
    }

    return p_astro



"""
Read and parse RapidPE output result for post-processing.
Author: Cory Chu <cory@gwlab.page>
"""

from functools import cached_property
from pathlib import Path
from joblib import Parallel, delayed
from configparser import ConfigParser
import re
import numpy as np
import h5py
# import pandas as pd
from .grid_point import RapidPE_grid_point
from .io import load_event_info_dict_txt
from .io import load_injection_info_txt
from .io import dict_of_ndarray_to_recarray
from .io import recarray_to_dict_of_ndarray
from .io import dict_from_hdf_group
from .io import dict_to_hdf_group

from .transform import transform_m1m2_to_mceta, transform_mceta_to_m1m2
from .transform import jacobian_mceta_by_m1m2
from .transform import Mass_Spin

from matplotlib.tri import Triangulation
from matplotlib.tri import LinearTriInterpolator, CubicTriInterpolator
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator
from scipy.interpolate import CloughTocher2DInterpolator
from scipy.stats import multinomial

# import time  # for profiling


def unique_with_tolerance(array, tolerance):
    tolerance = np.abs(tolerance)
    sorted_array = np.sort(array)
    diff = np.diff(sorted_array)
    mask = np.append(True, np.where(diff < tolerance, False, True))
    return sorted_array[mask]


def grid_separation_min(array):
    tolerance = (array.max() - array.min()) / np.sqrt(len(array)) / 200
    # FIXME: is 200 okay?
    # FIXME: is np.sqrt(len(array)) okay?
    x = unique_with_tolerance(array, tolerance=tolerance)
    return np.diff(x).min()


class RapidPE_result:
    """
    RapidPE Result

    It can be created from a list of *.xml.gz files using
        RapidPE_result.from_xml_array([a.xml.gz, b.xml.gz, ...])
    Check help(RapidPE_result.from_xml_array) for an example.

    ...

    Attributes
    ----------
    grid_points : [RapidPE_grid_point]
        list of all grid points data

    mass_i : numpy.ndarray
        array of mass_i over grid points

    spin_iz : numpy.ndarray
        array of spin_iz over grid points

    Similar cases for chirp_mass, symmetric_mass_ratio, ...

    Methods
    -------
    from_xml_array([a.xml.gz, b.xml.gz, ...]):
        Get result from xml.gz files

    do_interpolate_marg_log_likelihood_m1m2(method="cubic"):
        Perfom interpolation of marg_log_likelihood
        After executing, a new method log_likelihood(m1, m2)
        will be created.

    log_likelihood(m1, m2):
        Interpolated log_likelihood

    """

    def __init__(self, result=None):
        if result is None:
            self.grid_points = np.empty(0, dtype=object)
            self._keys = []
        else:
            self.grid_points = result.grid_points
            self._keys = result._keys

            for attr in result._keys + ["event_info",
                                        "injection_info",
                                        "config_info",
                                        "posterior_samples"]:
                try:
                    setattr(self, attr, getattr(result, attr))
                except AttributeError:
                    pass
            # Try:
            # self.mass_1 = result.mass_1
            # self.mass_2 = result.mass_2
            # self.spin_1z = result.spin_1z
            # self.spin_2z = result.spin_2z
            # self.marg_log_likelihood = result.marg_log_likelihood
            # ...

    def __getitem__(self, key):
        """
        Get res.attr by res["attr"]
        """
        return getattr(self, key)

    @cached_property
    def _x(self):
        x = Mass_Spin.from_m1m2(self.mass_1,
                                self.mass_2,
                                grid_coordinates=self.grid_coordinates)
        return x

    @cached_property
    def x1(self):
        return self._x.x1

    @cached_property
    def x2(self):
        return self._x.x2

    @property
    def grid_coordinates(self):
        """
        List of grid coordinates in wich the grid points are rectilinear
        """
        iparam_search = self.config_info["intrinsic_param_to_search"]
        distant_coord = self.config_info["distance_coordinates"]
        if iparam_search == "[mass1,mass2]":
            if distant_coord == "mchirp_q":
                return ["chirp_mass", "mass_ratio"]
            elif distant_coord == "mchirp_eta":
                return ["chirp_mass", "symmetric_mass_ratio"]
            else:
                raise NotImplementedError("distance_coordinates not supported")
        elif iparam_search == "[mass2,mass2,spin1z,spin2z]":
            if distant_coord == "mchirp_q":
                return ["chirp_mass", "mass_ratio", "chi_eff", "chi_a"]
            elif distant_coord == "mchirp_eta":
                return ["chirp_mass", "symmetric_mass_ratio", "chi_eff", "chi_a"]  # noqa E501
            else:
                raise NotImplementedError("distance_coordinates not supported")
        else:
            raise NotImplementedError("intrinsic_param_to_search not supported")  # noqa E501

    @cached_property
    def intrinsic_table(self):
        """
        Combine intrinsic tables to a single dict
        """
        # return pd.DataFrame({key: getattr(self, key) for key in self._keys})
        return {key: getattr(self, key) for key in self._keys}

    @cached_property
    def extrinsic_samples(self):
        """
        Combine extrinsic samples to a single dict
        """
        # return pd.concat(
        #      [pd.DataFrame(gp.extrinsic_table) for gp in self.grid_points])
        gps = self.grid_points
        keys_ext = gps[0].extrinsic_table.keys()
        extrinsic_samples = {}
        for key in keys_ext:
            extrinsic_samples[key] = \
                np.concatenate([gp.extrinsic_table[key] for gp in gps])
        return extrinsic_samples

    def __copy__(self):
        return RapidPE_result(self)

    def copy(self):
        return self.__copy__()

    @classmethod
    def from_hdf(cls, hdf_file, extrinsic_table=True):
        """
        Get result from a Rapid-PE HDF file

        Example
        -------
            result = RapidPE_result.from_hdf("path/to/hdf_file")

        Attributes
        ----------
        hdf_file : string
            The path to Rapid-PE HDF file

        extrinsic_table : bool
            Whether loading extrinsic_table as well

        """
        result = cls()
        with h5py.File(hdf_file, "r") as f:
            try:
                # Load grid_points
                gps = f["grid_points"]
                N = len(gps)
                result.grid_points = np.empty(N, dtype=object)
                for i, gp in enumerate(gps.values()):
                    result.grid_points[i] = \
                        RapidPE_grid_point.from_hdf_grid_point_group(
                            hdf_gp_group=gp,
                            extrinsic_table=extrinsic_table
                            )

                # Load intrinsic_table
                it = f["intrinsic_table"]
                result.intrinsic_table = {key: it[key] for key in it.dtype.names}  # noqa E501
                result._keys = list(it.dtype.names)
            except KeyError:
                pass

            # Load other attributes
            for attr in result._keys:
                try:
                    setattr(result, attr, result.intrinsic_table[attr])
                except KeyError:
                    pass

            # Load event_info and injection_info
            for attr in ["event_info", "injection_info", "config_info"]:
                try:
                    x = dict_from_hdf_group(f[attr])
                    setattr(result, attr, x)
                except KeyError:
                    pass

            # Load posterior samples
            try:
                result.posterior_samples = \
                     recarray_to_dict_of_ndarray(f["posterior_samples"])
            except KeyError:
                pass

        return cls(result)

    @classmethod
    def from_run_dir(cls,
                     run_dir,
                     use_numpy=True,
                     use_ligolw=True,
                     extrinsic_table=True,
                     load_results=True,
                     parallel_n=1):
        """
        Get result from a Rapid-PE run_dir

        Example
        -------
            result = RapidPE_result.from_run_dir("path/to/run_dir")

        Attributes
        ----------
        run_dir : string or pathlib's Path() object
            The path to rapid-pe run_dir

        use_ligolw : bool
            Whether using ligo.lw to read xml files

        extrinsic_table : bool
            Whether loading extrinsic_table as well

        """
        run_dir = Path(run_dir)

        if load_results:
            results_dir = Path(run_dir)/Path("results")
            xml_array = [f.as_posix() for f in sorted(results_dir.glob("*.xml.gz"))]  # noqa: E501

            result = cls.from_xml_array(xml_array,
                                        use_numpy=use_numpy,
                                        use_ligolw=use_ligolw,
                                        extrinsic_table=extrinsic_table,
                                        parallel_n=parallel_n)
        else:
            result = cls()

        event_info_dict_txt = run_dir / "event_info_dict.txt"
        injection_info_txt = run_dir / "injection_info.txt"
        config_ini = run_dir/"Config.ini"

        try:
            result.event_info = load_event_info_dict_txt(event_info_dict_txt)
        except FileNotFoundError:
            pass

        try:
            result.injection_info = load_injection_info_txt(injection_info_txt)
        except FileNotFoundError:
            pass

        try:
            config = ConfigParser()
            config.read(config_ini)
            result.config_info = \
                {"distance_coordinates":
                    config["GridRefine"]["distance-coordinates"],
                 "intrinsic_param_to_search":
                    config["General"]["intrinsic_param_to_search"]}
        except KeyError:
            pass

        return result

    @classmethod
    def from_xml_array(cls,
                       xml_array,
                       use_numpy=True,
                       use_ligolw=True,
                       extrinsic_table=True,
                       parallel_n=1):
        """
        Get result from xml.gz files

        Example:
            import glob
            results_dir = "path/to/results"
            result_xml_files = glob.glob(results_dir+"*.xml.gz")
            result = RapidPE_result.from_xml_array(result_xml_files)
        """
        result = cls()
        N = len(xml_array)
        # Get keys from the 1st xml file's intrinsic_table
        result._keys = \
            RapidPE_grid_point.from_xml(
                xml_array[0],
                use_numpy=use_numpy,
                use_ligolw=use_ligolw,
                extrinsic_table=extrinsic_table
            ).intrinsic_table.keys()
        result._keys = list(result._keys)

        # Initialize attributes
        for attr in result._keys:
            setattr(result, attr, np.zeros(N))

        # Initialize "grid_points" attribute
        result.grid_points = np.empty(N, dtype=object)

        # Setup for "iteration" attribute
        result._keys.append("iteration")
        result.iteration = np.zeros(N, dtype=int)
        re_iter = re.compile(r"ILE_iteration_([0-9]+).+\.xml\.gz")

        # Read XML grid points
        if parallel_n == 1:
            for i, filename in enumerate(xml_array):
                grid_point = RapidPE_grid_point.from_xml(
                    filename,
                    use_numpy=use_numpy,
                    use_ligolw=use_ligolw,
                    extrinsic_table=extrinsic_table
                )

                # Append grid-points
                result.grid_points[i] = grid_point

                # Append Intrinsic Parameters of grid-points
                for attr in result._keys:
                    try:
                        getattr(result, attr)[i] = \
                            grid_point.intrinsic_table[attr][0]
                    except KeyError:
                        pass

                # Append "iteration", i.e., grid-refinement-level
                result.iteration[i] = int(re_iter.search(filename).group(1))

        # Parallel Reading XML files
        else:
            # t0 = time.time()
            grid_points = Parallel(n_jobs=parallel_n)(
                delayed(RapidPE_grid_point.from_xml)(
                    filename,
                    use_numpy=use_numpy,
                    use_ligolw=use_ligolw,
                    extrinsic_table=extrinsic_table
                ) for filename in xml_array
                )
            # t1 = time.time()
            for i, grid_point in enumerate(grid_points):
                # Append grid-points
                result.grid_points[i] = grid_point

                # Append Intrinsic Parameters of grid-points
                for attr in result._keys:
                    try:
                        getattr(result, attr)[i] = \
                            grid_point.intrinsic_table[attr][0]
                    except KeyError:
                        pass

                # Append "iteration", i.e., grid-refinement-level
                filename = xml_array[i]
                result.iteration[i] = int(re_iter.search(filename).group(1))
            # t2 = time.time()
            # print("parallel reading time: ", t1-t0)
            # print("post-proccessing time: ", t2-t1)

        # Add chirp_mass and symmetric_mass_ratio
        if ("mass_1" in result._keys) and ("mass_2" in result._keys):
            x = Mass_Spin.from_m1m2(result.mass_1, result.mass_2)
            result.chirp_mass = x.chirp_mass
            result.symmetric_mass_ratio = x.symmetric_mass_ratio
            result.mass_ratio = x.mass_ratio
            result._keys.extend(
                ["chirp_mass", "symmetric_mass_ratio", "mass_ratio"])

        return cls(result)

    def to_hdf(self, hdf_filename, extrinsic_table=True, compression=None):
        """
        Save result to hdf file

        Parameters
        ----------
        filename : str
            The name of the hdf file

        compression : str, optional (default: None)
            The compression method, e.g., "gzip", "lzf".
        """

        with h5py.File(hdf_filename, 'w', track_order=True) as f:

            # Check if there are results
            if len(self.grid_points) != 0:

                # Check if there is extrinsic_table
                gp = self.grid_points[0]
                extrinsic_table &= len(gp.extrinsic_table) > 0

                # Create "grid_points" group to hold self.grid_points
                group_grid_points_raw = \
                    f.create_group("grid_points", track_order=True)

                for i, gp in enumerate(self.grid_points):
                    group_gp = group_grid_points_raw.create_group(str(i))

                    # Add intrinsic_table
                    it = dict_of_ndarray_to_recarray(gp.intrinsic_table)
                    group_gp.create_dataset("intrinsic_table", data=it)

                    # Add extrinsic_table
                    if extrinsic_table:
                        et = dict_of_ndarray_to_recarray(gp.extrinsic_table)
                        group_gp.create_dataset("extrinsic_table",
                                                data=et,
                                                compression=compression)

                    # Add xml_filename
                    group_gp.attrs["xml_filename"] = gp.xml_filename

                # Combine intrinsic parameters into "intrinsic_table" dataset
                result_np = dict_of_ndarray_to_recarray(self.intrinsic_table)
                f.create_dataset("intrinsic_table", data=result_np)

                # Create virtual dataset for "extrinsic_samples"
                if extrinsic_table:
                    gps = f["grid_points"]
                    ds0 = f["grid_points/0/extrinsic_table"]
                    n_samples = np.array(
                        [gp["extrinsic_table"].shape[0] for gp in gps.values()])  # noqa: E501
                    n_samples = np.cumsum(n_samples)
                    n_samples = np.concatenate([[0], n_samples])
                    layout = h5py.VirtualLayout(shape=(n_samples[-1], ),
                                                dtype=ds0.dtype)

                    for i, gp in enumerate(gps.values()):
                        vsource = h5py.VirtualSource(gp["extrinsic_table"])
                        layout[n_samples[i]:n_samples[i+1]] = vsource

                    f.create_virtual_dataset('extrinsic_samples', layout)

            # Save event_info, injection_info, and config_info
            for attr in ["event_info", "injection_info", "config_info"]:
                try:
                    x = getattr(self, attr)
                    g = f.create_group(attr)
                    dict_to_hdf_group(x, g)
                except AttributeError:
                    pass

            # Save posterior_samples
            try:
                data = dict_of_ndarray_to_recarray(self.posterior_samples)
                f.create_dataset("posterior_samples", data=data)
            except AttributeError:
                pass

    def do_interpolate_marg_log_likelihood_m1m2(
            self,
            method="linear-scipy",
            gaussian_sigma_to_grid_size_ratio=0.5
            ):
        """
        Perfom triangular interpolation of marg_log_likelihood in
        chirp_mass (M_c), symmetric_mass_ratio (eta) space.
        After executing, a new method log_likelihood(m1, m2)
        will be created.

        Parameters
        ----------
            method: str
                currently, we support "cubic", "linear",
                "linear-scipy", and "gaussian".

            gaussian_sigma_to_grid_size_ratio=0.5: float
                if using method="gaussian", we can change the
                sigma of gaussian with respect to grid size

        """

        _supported_methods = \
            'method= "cubic", "linear", "linear-scipy", "nearest-scipy",' \
            '"cubic-scipy", "gaussian", or "gaussian-renormalized"'

        grid_coord_1 = self[self.grid_coordinates[0]]
        grid_coord_2 = self[self.grid_coordinates[1]]

        if method == "gaussian" or method == "gaussian-renormalized":
            # def gaussian_log_likelihood(m1, m2):
            #     mc_arr, eta_arr = transform_m1m2_to_mceta(m1, m2)
            #     sigma_mc = grid_separation_min(self.chirp_mass) *\
            #         gaussian_sigma_to_grid_size_ratio
            #     sigma_eta = grid_separation_min(self.symmetric_mass_ratio) *\
            #         gaussian_sigma_to_grid_size_ratio

            #     likelihood = np.zeros_like(mc_arr)
            #     for i in range(len(self.chirp_mass)):
            #         likelihood += \
            #             np.exp(self.marg_log_likelihood[i]) * \
            #             np.exp(
            #                 (-0.5/sigma_mc**2
            #                  * (mc_arr - self.chirp_mass[i])**2) +
            #                 (-0.5/sigma_eta**2
            #                  * (eta_arr - self.symmetric_mass_ratio[i])**2)
            #             )
            #     return np.log(likelihood)
            def gaussian_log_likelihood(m1, m2):
                x = Mass_Spin.from_m1m2(m1, m2)
                x1 = x[self.grid_coordinates[0]]
                x2 = x[self.grid_coordinates[1]]

                grid_levels = np.unique(self.iteration)
                sigma_1 = {}
                sigma_2 = {}

                for gl in grid_levels:
                    sigma_1[gl] = grid_separation_min(
                            grid_coord_1[self.iteration == gl]
                        ) * gaussian_sigma_to_grid_size_ratio
                    sigma_2[gl] = grid_separation_min(
                            grid_coord_2[self.iteration == gl]
                        ) * gaussian_sigma_to_grid_size_ratio

                likelihood = np.zeros_like(x1)
                for i in range(len(grid_coord_1)):
                    likelihood += \
                        np.exp(self.marg_log_likelihood[i]) * \
                        np.exp(
                            (-0.5/sigma_1[self.iteration[i]]**2
                             * (x1 - grid_coord_1[i])**2) +
                            (-0.5/sigma_2[self.iteration[i]]**2
                             * (x2 - grid_coord_2[i])**2)
                        )
                return np.log(likelihood)
            self.log_likelihood = gaussian_log_likelihood

            if method == "gaussian-renormalized":
                # FIXME: The max point can shft if sigma is large.

                # old_max = np.exp(self.marg_log_likelihood.max())
                # old_max_idx = np.argmax(self.marg_log_likelihood)
                # new_max = np.exp(
                #     gaussian_log_likelihood(self.mass_1[old_max_idx],
                #                             self.mass_2[old_max_idx])
                # )

                self_interp = gaussian_log_likelihood(self.mass_1, self.mass_2)
                new_max = self_interp.max()
                new_max_idx = np.argmax(self_interp)
                old_max = self.marg_log_likelihood[new_max_idx]

                print("old_max: ", np.exp(old_max))
                print("new_max: ", np.exp(new_max))

                def gaussian_log_likelihood_renormalized(m1, m2):
                    return gaussian_log_likelihood(m1, m2) + old_max - new_max

                self.log_likelihood = gaussian_log_likelihood_renormalized

        elif method == "gaussian-numpy":
            """
            Gaussian approximation using numpy broadcasting.
            FIXME: For some reason, it can be slower than python for-loop...
            """
            def gaussian_log_likelihood(m1, m2):
                result = self
                mc_arr, eta_arr = transform_m1m2_to_mceta(m1, m2)
                sigma_mc = grid_separation_min(result.chirp_mass) *\
                    gaussian_sigma_to_grid_size_ratio
                sigma_eta = grid_separation_min(result.symmetric_mass_ratio) *\
                    gaussian_sigma_to_grid_size_ratio

                likelihood = np.sum(
                    np.exp(result.marg_log_likelihood) *
                    np.exp(
                        (-0.5/sigma_mc**2 *
                            (mc_arr[..., np.newaxis] - result.chirp_mass)**2) +
                        (-0.5/sigma_eta**2 *
                            (eta_arr[..., np.newaxis] -
                                result.symmetric_mass_ratio)**2)
                    ),
                    axis=-1
                )
                return np.log(likelihood)

            self.log_likelihood = gaussian_log_likelihood

        elif "scipy" in method:
            if method == "linear-scipy":
                f = LinearNDInterpolator(
                    # list(zip(self.chirp_mass, self.symmetric_mass_ratio)),
                    list(zip(grid_coord_1, grid_coord_2)),
                    self.marg_log_likelihood,
                    rescale=True,
                    fill_value=-100  # FIXME: is -100 okay?
                    )
            elif method == "cubic-scipy":
                f = CloughTocher2DInterpolator(
                    # list(zip(self.chirp_mass, self.symmetric_mass_ratio)),
                    list(zip(grid_coord_1, grid_coord_2)),
                    self.marg_log_likelihood,
                    rescale=True,
                    fill_value=-100  # FIXME: is -100 okay?
                    )
            elif method == "nearest-scipy":
                f = NearestNDInterpolator(
                    # list(zip(self.chirp_mass, self.symmetric_mass_ratio)),
                    list(zip(grid_coord_1, grid_coord_2)),
                    self.marg_log_likelihood,
                    rescale=True
                    )
            else:
                raise ValueError(_supported_methods)

            def log_likelihood(m1, m2):
                # mc, eta = transform_m1m2_to_mceta(m1, m2)
                # ll = f(mc, eta)
                x = Mass_Spin.from_m1m2(m1, m2)
                x1 = x[self.grid_coordinates[0]]
                x2 = x[self.grid_coordinates[1]]
                ll = f(x1, x2)
                return ll

            self.log_likelihood = log_likelihood

        else:
            raise DeprecationWarning(_supported_methods)
            # TODO: remove methods uisng matplotlib triangulation

            triangles = Triangulation(self.chirp_mass,
                                      self.symmetric_mass_ratio)

            if method == "cubic":
                f = CubicTriInterpolator(triangles, self.marg_log_likelihood)
            elif method == "linear":
                f = LinearTriInterpolator(triangles, self.marg_log_likelihood)
            else:
                raise ValueError(_supported_methods)

            def log_likelihood(m1, m2):
                # FIXME: if m1, m2 is not numpy.ndarray (e.g, scalar, it fails)
                mc, eta = transform_m1m2_to_mceta(m1, m2)
                ll = f(mc, eta)
                ll = np.ma.fix_invalid(ll, fill_value=-100).data
                # FIXME: is -100 okay?
                return ll

            self.log_likelihood = log_likelihood

    def do_interpolate_marg_log_likelihood_mass_spin():
        # TODO: do 4-D interpolation using
        #       scipy.interpolate.LinearNDInterpolator
        #       Ref:https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.LinearNDInterpolator.html
        raise NotImplementedError("to be support in the future")

    def generate_samples(self, N=5000, method="gaussian"):
        if method == "gaussian":
            # Grid-level-independent covariance matrix (old method)
            #
            # sigma_mc = grid_separation_min(self.chirp_mass) * 0.5
            # sigma_eta = grid_separation_min(self.symmetric_mass_ratio) * 0.5
            # cov = np.diag([sigma_mc**2, sigma_eta**2])

            # FIXME: use grid_coordinates instead of hard-coded
            # Compute covariance matrix for different grid levels
            grid_levels = np.unique(self.iteration)
            cov = {}
            for gl in grid_levels:
                mask = self.iteration == gl
                sigma_mc = grid_separation_min(
                    self.chirp_mass[mask]) * 0.5
                sigma_eta = grid_separation_min(
                    self.symmetric_mass_ratio[mask]) * 0.5
                cov[gl] = np.diag([sigma_mc**2, sigma_eta**2])

            # Compute likelihood at each grid point
            likelihood = np.exp(self.marg_log_likelihood)
            sum_likelihood = np.sum(likelihood)

            # Compute number of samples for each grid point
            N_multinomial = multinomial(N*20, likelihood/sum_likelihood)
            N_per_grid_point = N_multinomial.rvs(1)[0]

            # Generate samples
            samples = np.zeros([0, 2])
            for mc, eta, lh, gl, n in zip(self.chirp_mass,
                                          self.symmetric_mass_ratio,
                                          likelihood,
                                          self.iteration,
                                          N_per_grid_point):
                samples = np.concatenate([
                    samples,
                    np.random.multivariate_normal([mc, eta], cov[gl], n)
                    ])
            mc, eta = samples.T

            # Mask out the samples outside the grid region
            eta_max = self.symmetric_mass_ratio.max()
            eta_min = self.symmetric_mass_ratio.min()
            mask = np.logical_and(eta <= eta_max, eta >= eta_min)

            # mc_max = self.chirp_mass.max()
            # mc_min = self.chirp_mass.min()
            # mask = np.logical_and(mask, mc >= mc_min)
            # mask = np.logical_and(mask, mc <= mc_max)

            mc, eta = mc[mask], eta[mask]
            m1, m2 = transform_mceta_to_m1m2(mc, eta)

            prob = 1/jacobian_mceta_by_m1m2(m1, m2)
            prob /= np.sum(prob)

            m = [{"m1": m1[i], "m2": m2[i]} for i in range(len(m1))]
            # m_samples = np.random.choice(m, size=N, p=prob)
            m_samples = np.random.choice(m, size=N, p=prob, replace=False)
            m1, m2 = np.array([[x["m1"], x["m2"]] for x in m_samples]).T
            self.samples = {"mass_1": m1, "mass_2": m2}

    def generate_posterior_samples(self,
                                   N=5000,
                                   method="gaussian",
                                   gaussian_sigma_to_grid_size_ratio=1.0,
                                   em_bright_compatible=True):
        """
        Generate posterior samples.
        The generated samples are saved in self.samples

        Parameters
        ----------
        N: int
            Number of samples to generate
            Default: 5000

        method: str
            Method to generate samples
            Default: "gaussian"

        gaussian_sigma_to_grid_size_ratio: float
            Ratio of sigma to grid size
            Default: 1.0

        em_bright_compatible: bool
            If True, generate samples compatible with EM_BRIGHT
            Default: True

        """
        if method == "gaussian":
            grid_levels = np.unique(self.iteration)
            cov = {}
            for gl in grid_levels:
                mask = self.iteration == gl
                sigma_x1 = grid_separation_min(
                    self.x1[mask]) * gaussian_sigma_to_grid_size_ratio
                sigma_x2 = grid_separation_min(
                    self.x2[mask]) * gaussian_sigma_to_grid_size_ratio
                cov[gl] = np.diag([sigma_x1**2, sigma_x2**2])

            # Compute likelihood at each grid point
            likelihood = np.exp(self.marg_log_likelihood)
            sum_likelihood = np.sum(likelihood)

            # Compute number of samples for each grid point
            N_multinomial = multinomial(N*20, likelihood/sum_likelihood)
            N_per_grid_point = N_multinomial.rvs(1)[0]

            # Generate samples
            samples = np.zeros([0, 2])
            for x1, x2, lh, gl, n in zip(self.x1,
                                         self.x2,
                                         likelihood,
                                         self.iteration,
                                         N_per_grid_point):
                samples = np.concatenate([
                    samples,
                    np.random.multivariate_normal([x1, x2], cov[gl], n)
                    ])
            x1, x2 = samples.T

            # Mask out the samples outside the region of interest
            if self.grid_coordinates[0] == "chirp_mass":
                x1_min = 0.0
            else:
                raise ValueError("Unknown grid coordinate x1")

            if self.grid_coordinates[1] == "mass_ratio":
                x2_max = 1.0
                x2_min = 0.0
            elif self.grid_coordinates[1] == "symmetric_mass_ratio":
                x2_max = 0.25
                x2_min = 0.0
            else:
                raise ValueError("Unknown grid coordinate x2")

            mask = x2 <= x2_max
            mask &= x2 > x2_min
            mask &= x1 > x1_min

            x1, x2 = x1[mask], x2[mask]

            x = Mass_Spin.from_x1x2(x1,
                                    x2,
                                    grid_coordinates=self.grid_coordinates)

            # An ad-hoc cut on the mass_1 and mass_2
            mask_m1m2 = x.mass_1 < 500
            mask_m1m2 &= x.mass_2 > 0.1

            # Re-weight the samples according to the Jacobian such that
            # it has a uniform prior in m1-m2 space
            # Reference: https://dcc.ligo.org/LIGO-T2300198
            weight = x.jacobian_m1m2_by_x1x2[mask_m1m2]
            weight /= np.sum(weight)

            m = {"mass_1": x.mass_1[mask_m1m2],
                 "mass_2": x.mass_2[mask_m1m2]}

            m = dict_of_ndarray_to_recarray(m)
            samples = np.random.choice(m, size=N, p=weight, replace=False)
            samples = recarray_to_dict_of_ndarray(samples)

            if em_bright_compatible:
                shape = samples["mass_1"].shape
                event_spin = self.event_info["event_spin"]
                spin_1z = np.broadcast_to(event_spin["spin_1z"], shape)
                spin_2z = np.broadcast_to(event_spin["spin_2z"], shape)

                samples["mass_1_source"] = samples["mass_1"]
                samples["mass_2_source"] = samples["mass_2"]
                samples["spin_1z"] = spin_1z
                samples["spin_2z"] = spin_2z

                self.posterior_samples = samples
            else:
                self.posterior_samples = samples



"""
Utility functions of coordinate transformation.
"""

import numpy as np


class Mass_Spin:
    def __init__(self, obj=None):
        if obj is None:
            self._mass_1 = None
            self._mass_2 = None
            self._spin_1z = None
            self._spin_2z = None
            self.grid_coordinates = None
        else:
            self._mass_1 = obj._mass_1
            self._mass_2 = obj._mass_2
            self._spin_1z = obj._spin_1z
            self._spin_2z = obj._spin_2z
            self.grid_coordinates = obj.grid_coordinates

    def __getitem__(self, key):
        """
        Get obj.attr by obj["attr"]
        """
        return getattr(self, key)

    @classmethod
    def from_x1x2(cls, x1, x2, grid_coordinates):
        """
        Create a Mass_Spin object from the first two mass-spin coordinates
        x1, x2 defined in the "grid_coordinates"

        Parameters
        ----------
        x1: float or np.ndarray
            The first mass-spin coordinate
        x2: float or np.ndarray
            The second mass-spin coordinate
        grid_coordinates: list
            The two mass-spin coordinates defining the grid
            e.g.
                ["chirp_mass", "mass_ratio"]
                ["chirp_mass", "symmetric_mass_ratio"]
                ["chirp_mass", "mass_ratio", "chi_eff", "chi_a"]
                ["chirp_mass", "symmetric_mass_ratio", "chi_eff", "chi_a"]

        Returns
        -------
        obj: Mass_Spin
            The Mass_Spin object

        """
        obj = cls()
        obj.grid_coordinates = grid_coordinates
        if grid_coordinates[0] == "chirp_mass":
            if grid_coordinates[1] == "mass_ratio":
                obj._mass_1, obj._mass_2 = cls.mcq_to_m1m2(x1, x2)
            elif grid_coordinates[1] == "symmetric_mass_ratio":
                obj._mass_1, obj._mass_2 = cls.mceta_to_m1m2(x1, x2)
            else:
                raise ValueError("Invalid grid_coordinates")
        else:
            raise ValueError("Invalid grid_coordinates")
        return obj

    @classmethod
    def from_m1m2s1zs2z(cls,
                        mass_1,
                        mass_2,
                        spin_1z,
                        spin_2z,
                        grid_coordinates=None):
        obj = cls()
        obj._mass_1 = mass_1
        obj._mass_2 = mass_2
        obj._spin_1z = spin_1z
        obj._spin_2z = spin_2z
        obj.grid_coordinates = grid_coordinates
        return obj

    @classmethod
    def from_m1m2(cls, mass_1, mass_2, grid_coordinates=None):
        obj = cls()
        obj._mass_1 = mass_1
        obj._mass_2 = mass_2
        obj.grid_coordinates = grid_coordinates
        return obj

    @classmethod
    def from_mceta(cls, mc, eta, grid_coordinates=None):
        obj = cls()
        m1, m2 = cls.mceta_to_m1m2(mc, eta)
        obj._mass_1 = m1
        obj._mass_2 = m2
        obj.grid_coordinates = grid_coordinates
        return obj

    @classmethod
    def from_mcq(cls, mc, q, grid_coordinates=None):
        obj = cls()
        m1, m2 = cls.mcq_to_m1m2(mc, q)
        obj._mass_1 = m1
        obj._mass_2 = m2
        obj.grid_coordinates = grid_coordinates
        return obj

    @property
    def x1x2(self):
        x1 = self[self.grid_coordinates[0]]
        x2 = self[self.grid_coordinates[1]]
        return x1, x2

    @property
    def x1(self):
        x1 = self[self.grid_coordinates[0]]
        return x1

    @property
    def x2(self):
        x2 = self[self.grid_coordinates[1]]
        return x2

    @property
    def mass_1(self):
        return self._mass_1

    @mass_1.setter
    def mass_1(self, value):
        self._mass_1 = value

    @property
    def mass_2(self):
        return self._mass_2

    @mass_2.setter
    def mass_2(self, value):
        self._mass_2 = value

    @property
    def chirp_mass(self):
        chirp_mass = self.m1m2_to_mc(self._mass_1, self._mass_2)
        return chirp_mass

    @property
    def mass_ratio(self):
        mass_ratio = self.m1m2_to_q(self._mass_1, self._mass_2)
        return mass_ratio

    @property
    def symmetric_mass_ratio(self):
        symmetric_mass_ratio = self.m1m2_to_eta(self._mass_1, self._mass_2)
        return symmetric_mass_ratio

    @property
    def mceta(self):
        return self.chirp_mass, self.symmetric_mass_ratio

    @property
    def m1m2(self):
        return self._mass_1, self._mass_2

    @property
    def mcq(self):
        return self.chirp_mass, self.mass_ratio

    @property
    def jacobian_x1x2_by_m1m2(self):
        if self.grid_coordinates[0] == "chirp_mass":
            if self.grid_coordinates[1] == "mass_ratio":
                return self.jacobian_mcq_by_m1m2(self._mass_1, self._mass_2)
            elif self.grid_coordinates[1] == "symmetric_mass_ratio":
                return self.jacobian_mceta_by_m1m2(self._mass_1, self._mass_2)
            else:
                raise ValueError("Invalid grid_coordinates")
        else:
            raise ValueError("Invalid grid_coordinates")

    @property
    def jacobian_m1m2_by_x1x2(self):
        return 1/self.jacobian_x1x2_by_m1m2

    @classmethod
    def _norm_sym_ratio(cls, eta):
        """
        Copied from https://git.ligo.org/rapidpe-rift/rapidpe/-/blob/master/rapid_pe/lalsimutils.py
        """  # noqa E501
        # Assert phyisicality
        assert np.all(eta <= 0.25)
        return np.sqrt(1. - 4. * eta)

    @classmethod
    def mceta_to_m1m2(cls, mc, eta):
        """
        Compute component masses from Mc, eta. Returns m1 >= m2
        """
        m1 = 0.5*mc*eta**(-0.6)*(1. + cls._norm_sym_ratio(eta))  # 3/5=0.6
        m2 = 0.5*mc*eta**(-0.6)*(1. - cls._norm_sym_ratio(eta))  # 3/5=0.6
        return m1, m2

    @classmethod
    def m1m2_to_mceta(cls, m1, m2):
        """
        Compute chirp mass and symmetric mass ratio from component masses
        """
        mc = cls.m1m2_to_mc(m1, m2)
        eta = cls.m1m2_to_eta(m1, m2)
        return mc, eta

    @classmethod
    def mcq_to_m1m2(cls, mc, q):
        m = mc * (1+q)**0.2  # 1/5=0.2
        m1 = m / q**0.6  # 3/5=0.6
        m2 = m * q**0.4  # 2/5=0.4
        return m1, m2

    @classmethod
    def m1m2_to_mcq(cls, m1, m2):
        """
        Compute chirp mass and mass ratio from component masses

        Note:
            mass ratio q=m2/m1
            forcing m1 >= m2
        """
        mc = cls.m1m2_to_mc(m1, m2)
        q = cls.m1m2_to_q(m1, m2)
        return mc, q

    @classmethod
    def m1m2_to_q(cls, m1, m2):
        """
        Compute mass ratio from component masses
        q = m2/m1, where m1 >= m2
        """
        q = np.minimum(m1, m2) / np.maximum(m1, m2)
        return q

    @classmethod
    def m1m2_to_mc(cls, m1, m2):
        """
        Compute chirp mass from component masses
        """
        mc = (m1*m2)**0.6 * (m1+m2)**(-0.2)  # 3/5=0.6, 1/5=0.2
        return mc

    @classmethod
    def m1m2_to_eta(cls, m1, m2):
        """
        Compute symmetric mass ratio from component masses
        """
        eta = m1*m2 / (m1+m2)**2
        return eta

    @classmethod
    def jacobian_m1m2_by_mceta(cls, mc, eta):
        """
        Return the Jacobian (m1, m2)/(mc, eta)
        """
        assert np.all(mc > 0), "chirp_mass (Mc) should > 0"
        assert np.all(eta > 0), "symmetric_mass_ratio (eta) should > 0"
        assert np.all(eta <= 0.25), "symmetric_mass_ratio (eta) should <= 0.25"

        return mc / (np.sqrt(1. - 4.*eta) * eta**1.2)  # 6/5 = 1.2

    @classmethod
    def jacobian_mceta_by_m1m2(cls, m1, m2):
        """
        Return the Jacobian (mc, eta)/(m1, m2)
        """
        return (m1-m2)*(m1*m2)**0.6 / (m1+m2)**3.2  # 3/5 = 0.6, 16/5 = 3.2

    @classmethod
    def jacobian_m1m2_by_mcq(cls, mc, q):
        """
        Return the Jacobian (m1, m2)/(mc, q)
        """
        assert np.all(mc > 0), "chirp_mass (Mc) should > 0"
        assert np.all(q > 0), "mass_ratio (q) should > 0"

        return mc * (1+q)**0.4 / q**1.2  # 2/5 = 0.4, 6/5 = 1.2

    @classmethod
    def jacobian_mcq_by_m1m2(cls, m1, m2):
        """
        Return the Jacobian (mc, q)/(m1, m2)
        """
        return (m1*m2)**0.6 / (m1**2 * (m1+m2)**0.2)  # 3/5 = 0.6, 1/5 = 0.2


# =====================================================================
# For legacy support
# =====================================================================


def transform_m1m2_to_mceta(m1, m2):
    return Mass_Spin.m1m2_to_mceta(m1, m2)


def transform_mceta_to_m1m2(mc, eta):
    return Mass_Spin.mceta_to_m1m2(mc, eta)


def jacobian_m1m2_by_mceta(mc, eta):
    return Mass_Spin.jacobian_m1m2_by_mceta(mc, eta)


def jacobian_mceta_by_m1m2(m1, m2):
    return Mass_Spin.jacobian_mceta_by_m1m2(m1, m2)


# =====================================================================



"""
Parse the RapidPE results


!!!! IMPORTANT !!!!
!!!! IMPORTANT !!!!
!!!! IMPORTANT !!!!

    All classes in this file (parser.py) is deprecated.
    Please use RapidPE_result.from_xml_array() defined in result.py instead.

!!!! IMPORTANT !!!!
!!!! IMPORTANT !!!!
!!!! IMPORTANT !!!!


Author: Cory Chu <cory@gwlab.page>
"""

# For RapidPE_XML_fast
import xml.etree.ElementTree as ET
import gzip

# For RapidPE_XML
from ligo.lw import utils, lsctables, ligolw
lsctables.use_in(ligolw.LIGOLWContentHandler)


class RapidPE_XML:
    """
    Parse the RapidPE result ligo-lw XML files

    ...

    Attributes
    ----------
    TODO: Complete the example/explanation of parameters
    intrinsic_table : dict
        Intrinsic parameters get from "sngl_inspiral:table"
        with canonical parameter names.

        An example:
        {
            "mass_1": [1.4],
            "mass_2": [2.8],
            ...
        }

        Parameters:
        mass_i:
            component masses (in solar mass?)

        spin_iz:
            z-component of spin (what coordinate?)

        marg_log_likelihood:
            natural log of the likelihood
            marginalized over extrinsic parameters

        tau0, tau3:
            chirptime parameters
            It's a kind of mass parameter
            like (m1, m2) or (chirp_mass, mass_ratio)
            See Appendix B of https://arxiv.org/pdf/0706.4437.pdf



    extrinsic_table : dict
        Extrinsic parameters get from "sim_inspiral:table"
        with canonical parameter names


    Attributes (Advanced usage)
    ----------
    xmldoc : ligo.lw.ligolw.Document
        The opened ligolw xml file

    intrinsic_table_raw : dict
        Raw intrinsic parameters get from "sngl_inspiral:table"

    extrinsic_table_raw : dict
        Raw extrinsic parameters get from "sim_inspiral:table"

    """
    def __init__(self, filename: str):
        self.xmldoc = self._open_ligolw(filename)

        # Extract XML, assign to "raw" attributes
        # "sngl_inspiral:table" -> self.intrinsic_table_raw
        # "sim_inspiral:table" -> self.extrinsic_table_raw
        self.intrinsic_table_raw = self._get_ligolw_table(
            lsctables.SnglInspiralTable
            )
        try:
            self.extrinsic_table_raw = self._get_ligolw_table(
                lsctables.SimInspiralTable
                )
        except ValueError:
            # if not *samples.xml.gz or new version xml,
            # there is no "sim_inspiral:table"
            self.extrinsic_table_raw = {}
            pass

        # Provide canonical interface with two attributes
        #   self.intrinsic_table
        #   self.extrinsic_table
        # =============================================
        # TODO: Check the naming convention in bilby and PESummary
        #       [bilby]    https://arxiv.org/pdf/2006.00714.pdf
        #       [PESummary]https://docs.ligo.org/lscsoft/pesummary/stable_docs/gw/parameters.html
        #       spin1z = spin_1z(PESummary)? (maybe = chi_1x= spin_1_z (bilby))
        #       spin2z = spin_2z(PESummary)? (maybe = chi_2 = spin_2_z (bilby))
        #       distance = luminosity_distance?
        #       latitude = dec ?
        #       longitude = ra ?
        #       inclination = theta_jn ?
        #       polarization = psi?
        #       sampling_function is
        #         the p_s in Eq. 28 of https://arxiv.org/pdf/1502.04370.pdf
        # =============================================

        # Maps are "rapidpe_name": "canonical_name"
        intrinsic_parameter_map = {
                                    "mass1": "mass_1",
                                    "mass2": "mass_2",
                                    "spin1z": "spin_1z",
                                    "spin2z": "spin_2z",
                                    "snr": "marg_log_likelihood",
                                    "tau0": "tau0",
                                    "tau3": "tau3",
                                    "alpha4": "eccentricity",
                                }
        self.intrinsic_table = {}
        for key in intrinsic_parameter_map:
            try:
                self.intrinsic_table[intrinsic_parameter_map[key]] \
                    = self.intrinsic_table_raw[key]
            except KeyError:
                pass

        extrinsic_parameter_map = {
                                    "mass1": "mass_1",
                                    "mass2": "mass_2",
                                    "distance": "luminosity_distance",
                                    "latitude": "dec",
                                    "longitude": "ra",
                                    "inclination": "iota",
                                    "polarization": "psi",
                                    "alpha1": "log_likelihood",
                                    "alpha2": "prior",
                                    "alpha3": "sampling_function",
                                }
        self.extrinsic_table = {}
        for key in extrinsic_parameter_map:
            try:
                self.extrinsic_table[extrinsic_parameter_map[key]] \
                    = self.extrinsic_table_raw[key]
            except KeyError:
                pass

    def _open_ligolw(self, xmlfile: str):
        xmldoc = utils.load_filename(
            xmlfile, contenthandler=ligolw.LIGOLWContentHandler
            )
        return xmldoc

    def _get_ligolw_table(self, lsctable):
        xmltable = lsctable.get_table(self.xmldoc)
        table = {key: [] for key in xmltable.columnnames}
        for row in xmltable:
            for key in table.keys():
                table[key].append(getattr(row, key))
        return table


class RapidPE_XML_fast:
    """
    Faster Parser than RapidPE_XML

    ...

    """
    def __init__(self, filename: str):

        # Extract XML, assign to "raw" attributes
        # "sngl_inspiral:table" -> self.intrinsic_table_raw
        # "sim_inspiral:table" -> self.extrinsic_table_raw
        input_xml_gz = filename

        self.intrinsic_table_raw = self._get_ligolw_table(
            input_xml_gz,
            tablename="sngl_inspiral:table"
            )
        self.extrinsic_table_raw = self._get_ligolw_table(
            input_xml_gz,
            tablename="sim_inspiral:table"
            )

        # Provide canonical interface with two attributes
        #   self.intrinsic_table
        #   self.extrinsic_table

        # Maps are "rapidpe_name": "canonical_name"
        intrinsic_parameter_map = {
                                    "mass1": "mass_1",
                                    "mass2": "mass_2",
                                    "spin1z": "spin_1z",
                                    "spin2z": "spin_2z",
                                    "snr": "marg_log_likelihood",
                                    "tau0": "tau0",
                                    "tau3": "tau3",
                                    "alpha4": "eccentricity",
                                }
        self.intrinsic_table = {}
        for key in intrinsic_parameter_map:
            try:
                self.intrinsic_table[intrinsic_parameter_map[key]] \
                    = self.intrinsic_table_raw[key]
            except KeyError:
                pass

        extrinsic_parameter_map = {
                                    "mass1": "mass_1",
                                    "mass2": "mass_2",
                                    "distance": "luminosity_distance",
                                    "latitude": "dec",
                                    "longitude": "ra",
                                    "inclination": "iota",
                                    "polarization": "psi",
                                    "alpha1": "log_likelihood",
                                    "alpha2": "prior",
                                    "alpha3": "sampling_function",
                                }
        self.extrinsic_table = {}
        for key in extrinsic_parameter_map:
            try:
                self.extrinsic_table[extrinsic_parameter_map[key]] \
                    = self.extrinsic_table_raw[key]
            except KeyError:
                pass

    # Private Methods
    def _get_root_xml_gz(self, input_xml_gz):
        input_xml = gzip.open(input_xml_gz, 'r')
        tree = ET.parse(input_xml)
        root = tree.getroot()
        return root

    def _append_rows(self, keys, rows):
        result = {key: [] for key in keys}
        for row in rows.splitlines():
            row = row.split(',')
            # For some reason, some table have extra trailing ","
            # That makes an extra empty column in each row
            # So, we only keep first len(keys) columns
            row = row[0:len(keys)]
            for i, key in enumerate(keys):
                result[key].append(float(row[i]))
        return result

    def _get_ligolw_table(
            self,
            input_xml_gz,
            tablename="sngl_inspiral:table"
            ):

        # Get XML root
        root = self._get_root_xml_gz(input_xml_gz)

        # Find inspiral Table
        inspiral_table = root.find("*[@Name='" + tablename + "']")

        # Find coloum names
        keys = [i.attrib["Name"].split(":")[-1]
                for i in inspiral_table.iter("Column")]

        # Clean up string
        # TODO: Check reliablity! This is a dirty way to clean up the ligolwXML
        s = inspiral_table.find("Stream").text
        s = s.replace("\t", "")
        s = s[1:]  # Remove the extra "\n" in the begining of Stream string

        # Combine key-value
        return self._append_rows(keys, s)



from .grid_point import RapidPE_grid_point
from .result import RapidPE_result
from .p_astro import p_astro
from .plot import pretty_plt
from .transform import Mass_Spin

__all__ = ["RapidPE_grid_point",
           "RapidPE_result",
           "Mass_Spin",
           "p_astro",
           "pretty_plt"]
